/*
*
*    sleep.c - plots a sleep diary
*    version 0.7
*    (c) 2001 Steve G Parker (steve@steve-parker.org)
*    
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation version 2
* of the License.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* With thanks to Peter (peter-is-my-name@home.com) for sorting me out on
* Variance, and the need to divide it by 60^2.
*

  Files Required: diary.dat, sleep.dat, other.dat

  diary.dat
  eg: Fri 11 May 2
  	Means: "Friday 11th May - overall feeling is 2/5"

  sleep.dat
  eg: 26 Jun 00 45 26 Jun 04 00 26 Jun 09 15   
  ie: dd mmm hh mm x3
  1st = Try to sleep
  2nd = Get to sleep
  3rd = Wake up

  other.dat
  eg: drink 23 Jul 23 00 24 Jul 07 00 comment
  eg: 95work 23 Jul 09 00 23 Jul 17 30
  eg: work 23 Jul 12 00 23 Jul 18 30
  Or: type dd mmm hh mm dd mmm hh mm
  Where: type is the type of activity (and gif/type.gif exists)
       dd mmm is the date
       hh mm is the start (or finish) time of the activity.
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

FILE *statsfile;
int timebar;
int short_output;
int comments;
int short_length;
int details;
int alcohol;
int work;
FILE *workfile;

struct time_struct
{
  int h;
  int m;
};
struct date_struct
{
  char dom[5];
  int d;
  char mon[5];
  int m;
};
struct tod_struct
{
  struct date_struct d;
  struct time_struct time;
};
struct sleep_struct
{
  struct tod_struct try;
  struct tod_struct sleep;
  struct tod_struct wake;
  struct sleep_struct *next;
  struct sleep_struct *prev;
};
struct other_struct
{
  char *name;
  struct tod_struct start;
  struct tod_struct finish;
  struct other_struct *next;
  struct other_struct *prev;
  char *comment;
};
struct diary_struct
{
  struct date_struct date;
  int overall;
  int qos;
  char *feel;
  char *comments;
  struct sleep_struct *sleep;
  struct other_struct *other;
  struct diary_struct *next;
  struct diary_struct *prev;
  char state[20];
  int work;
  char work_descr[100];
};

struct week_struct
{
  struct diary_struct *first_day;
  struct week_struct *next;
  struct week_struct *prev;
  int work;
  int sleep;
  int idx;
};

#define RATIO 0.33
#define MAXENTRIES 1000
// Is this necessary? It's just for the stats array...
#define ACT_TRY 0		// These are for the
#define ACT_SLEEP 1		// stats array
#define ACT_WAKE 2

int stats[3][MAXENTRIES];
struct diary_struct *first_entry;
struct week_struct *first_week;
struct week_struct *thisweek;
char summary[1024];
char tmp_summ[255];
int sleep_today;
FILE *html;


void krqsort (int v[], int left, int right)
{
/* qsort from K&R; renamed from qsort as glibc already has a qsort */
  int i, last, x, y;


  if (left >= right)
    return;
  x = (left + right) / 2;
  y = v[left];
  v[left] = v[x];
  v[x] = y;
  last = left;
  for (i = left + 1; i <= right; i++)
    if (v[i] < v[left]) {
      last++;
      y = v[last];
      v[last] = v[i];
      v[i] = y;
    }
  y = v[left];
  v[left] = v[last];
  v[last] = y;
  krqsort (v, left, last - 1);
  krqsort (v, last + 1, right);
}

void sort_stats (int idx)
{
  int i;

  for (i = 0; i < 3; i++) {
    krqsort (stats[i], 0, idx);
  }
}


void m2mh (int min, char *s)
{
  int m, h;

  m = (min % 60);
  h = (min / 60);

  sprintf (s, "%02d:%02d", h, m);
}

void show_stats (int idx, struct week_struct *thisweek)
{
  int i, j, c, mode;
  char t[10];
  int mean;
  double variance = 0.0;
  double standarddeviation = 0.0;
  int totMin, totMax, totMean, totMode = 0;

/*
  // This shows the raw data, be it sorted or unsorted.
  fprintf(html, "<table border=1>\n");
  for (i=0; i<3; i++) {
    fprintf(html, "<tr>");
    for (j=0; j<=idx; j++) {
      fprintf(html, "<td>%d</td>", stats[i][j]);
    }
    fprintf(html, "</tr>\n");
  }
  fprintf(html, "</table>\n");
*/

  fprintf (statsfile,
	   "<tr><th>wc <a href=\"sleep-diary-wc-%s.%d.%s.html\">%s %d %s</a></th><th>Min</th><th>Max</th><th>Mean</th><th>Mode</th><th>Variance</th><th>SD</th></tr>\n",
	   thisweek->first_day->date.dom, thisweek->first_day->date.d, thisweek->first_day->date.mon,
	   thisweek->first_day->date.dom, thisweek->first_day->date.d, thisweek->first_day->date.mon);
  fprintf (html, "<table border=1>\n");
  fprintf (html,
	   "<tr><th>Stats (%d days)</th><th>Min</th><th>Max</th><th>Mean</th><th>Mode</th><th>Variance</th><th>SD</th></tr>\n",
	   idx);

  i = 0;
  while (stats[ACT_TRY][i] == 0)
    i++;
  m2mh (stats[ACT_TRY][i], t);
  fprintf (html, "<tr><th>Trying</th><td>%s</td>", t);
  fprintf (statsfile, "<tr><th>Trying</th><td>%s</td>", t);
  totMin = stats[ACT_TRY][i];
  m2mh (stats[ACT_TRY][idx], t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  totMax = stats[ACT_TRY][idx];
  // calc mean here
  c = 0;
  for (i = 1; i <= idx; i++) {
    c += stats[ACT_TRY][i];
  }
  c = (c / idx);
  mean = c;
  m2mh (c, t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  totMean = c;
  // calc mode here
  if ((idx % 2) == 1)
    mode = stats[ACT_TRY][(idx / 2) + 1];
  else {
    c = idx / 2;		// this rounds up...
    mode = (stats[ACT_TRY][c] + stats[ACT_TRY][c + 1]) / 2;
  }
  m2mh (mode, t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  totMode = mode;


/* This is the start of the Variance and Std Deviation bit
   SD is the sqrt of the Variance.
http://davidmlane.com/hyperstat/A16252.html
http://davidmlane.com/hyperstat/A40397.html
 
*/
  variance = 0.0;
  for (i = 1; i < idx; i++) {
    int v = 0;
    v = stats[ACT_TRY][i] - mean;
    v = v * v;
    variance += v;
  }
  variance = variance / (idx * 3600);
  //printf("variance=%f\n", variance);
  m2mh (variance, t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  m2mh (sqrt (variance), t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
/* This is the end of the Variance and Std Deviation bit */

  i = 0;
  while (stats[ACT_SLEEP][i] == 0)
    i++;
  m2mh (stats[ACT_SLEEP][i], t);
  fprintf (html, "<tr><th>Sleeping</th><td>%s</td>", t);
  fprintf (statsfile, "<tr><th>Sleeping</th><td>%s</td>", t);
  totMin += stats[ACT_SLEEP][i];
  m2mh (stats[ACT_SLEEP][idx], t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  totMax += stats[ACT_SLEEP][idx];
  // calc mean here
  c = 0;
  for (i = 1; i <= idx; i++) {
    c += stats[ACT_SLEEP][i];
  }
  c = (c / idx);
  mean = c;
  m2mh (c, t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  totMean += c;
  if ((idx % 2) == 1)
    mode = stats[ACT_SLEEP][(idx / 2) + 1];
  else {
    c = idx / 2;		// this rounds up...
    mode = (stats[ACT_SLEEP][c] + stats[ACT_SLEEP][c + 1]) / 2;
  }
  m2mh (mode, t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  totMode += mode;

  variance = 0.0;
  for (i = 1; i < idx; i++) {
    int v = 0;
    v = stats[ACT_SLEEP][i] - mean;
    v = v * v;
    variance += v;
  }
  variance = variance / (idx * 3600);
  //printf("variance=%f\n", variance);
  m2mh (variance, t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  m2mh (sqrt (variance), t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  fprintf (html, "</tr>\n");
  fprintf (statsfile, "</tr>\n");


  i = 0;
  while (stats[ACT_WAKE][i] == 0)
    i++;
  m2mh (stats[2][i], t);
  fprintf (html, "<tr><th>Awake</th><td>%s</td>", t);
  fprintf (statsfile, "<tr><th>Awake</th><td>%s</td>", t);
  totMin += stats[2][i];
  m2mh (stats[2][idx], t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  totMax += stats[2][idx];
  // calc mean here
  c = 0;
  for (i = 1; i <= idx; i++) {
    c += stats[ACT_WAKE][i];
  }
  c = (c / idx);
  m2mh (c, t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  totMean += c;
  if ((idx % 2) == 1)
    mode = stats[ACT_WAKE][(idx / 2) + 1];
  else {
    c = idx / 2;		// this rounds up...
    mode = (stats[ACT_WAKE][c] + stats[ACT_WAKE][c + 1]) / 2;
  }
  m2mh (mode, t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  totMode += mode;

  variance = 0.0;
  for (i = 1; i < idx; i++) {
    int v = 0;
    v = stats[ACT_WAKE][i] - mean;
    v = v * v;
    variance += v;
  }
  variance = variance / (idx * 3600);
  //printf("variance=%f\n", variance);
  m2mh (variance, t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  m2mh (sqrt (variance), t);
  fprintf (html, "<td>%s</td>\n", t);
  fprintf (statsfile, "<td>%s</td>\n", t);
  fprintf (html, "</tr>\n");
  fprintf (statsfile, "</tr>\n");

  m2mh (totMin, t);
  fprintf (html, "<tr><th>Totals</th><td>%s</td>", t);
  fprintf (statsfile, "<tr><th>Totals</th><td>%s</td>", t);
  m2mh (totMax, t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  m2mh (totMean, t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  m2mh (totMode, t);
  fprintf (html, "<td>%s</td>", t);
  fprintf (statsfile, "<td>%s</td>", t);
  fprintf (html, "<td>&nbsp;</td><td>&nbsp;</td></tr>", totMin, totMax, totMean);
  fprintf (statsfile, "<td>&nbsp;</td><td>&nbsp;</td></tr>", totMin, totMax, totMean);
  fprintf (html, "</table>\n");

}

int same_day (struct date_struct d1, struct date_struct d2)
{
  if ((d1.d == d2.d) && (!strcmp (d1.mon, d2.mon))) {
    return 1;
  }
  else {
    return 0;
  }
}

void show_sunrise (struct diary_struct *d)
{
  struct other_struct *o;
  int m1, m2, width;

  o = d->other;

  while (o != NULL) {
    if (!strcmp (o->name, "sunrise"))
      if (same_day (o->start.d, d->date))
	break;
    o = o->next;
  }

  if (o == NULL)
    return;

  if (!strcmp (o->name, "sunrise")) {
    m1 = o->start.time.h * 60 + o->start.time.m;
    width = m1 * RATIO;
    if (m1 > 0)
      fprintf (html, "<BR><img border=0 src=\"gif/trans.gif\" alt=\"\" width=\"%d\" height=\"3\">", width);
    m2 = o->finish.time.h * 60 + o->finish.time.m;
    m2 = m2 - m1;
    width = m2 * RATIO;
    fprintf (html,
	     "<img border=1 src=\"gif/%s.gif\" alt=\"sunrise %02d:%02d - %02d:%02d\" width=\"%d\" height=\"3\">",
	     o->name, o->start.time.h, o->start.time.m, o->finish.time.h, o->finish.time.m, width);
    //fprintf (html, "<BR>\n");
    //printf ("Sunrise: %02d %s %02d:%02d - %02d:%02d\n", d->date.d, d->date.mon, o->start.time.h,
    //o->start.time.m, o->finish.time.h, o->finish.time.m);
  }
}

void show_other (struct other_struct *o, struct diary_struct *d)
{
  int width, m1, m2;

  if ((!same_day (o->start.d, d->date))
      && (!same_day (o->finish.d, d->date)))
    return;

  if ((alcohol == 0) && (!strcmp (o->name, "drink")))
    return;

  if (!strcmp (o->name, "sunrise"))
    return;

/*
	Don't want to add text on non-sleep items.
  if (same_day(o->start.d, d->date)) 
    strcat(summary, o->comment);
*/

  if (same_day (o->start.d, d->date)) {
    m1 = o->start.time.h * 60 + o->start.time.m;
    width = m1 * RATIO;
    if (m1 > 0) {
      if ((o->prev == NULL) || (strcmp (o->prev->name, o->name)))
	fprintf (html, "<img border=0 src=\"gif/trans.gif\" alt=\"\" width=\"%d\" height=\"3\">", width);
    }
  }
  else
    m1 = 0;
  if (same_day (o->finish.d, d->date))
    m2 = o->finish.time.h * 60 + o->finish.time.m;
  else
    m2 = 24 * 60;
  width = m2 - m1;
  width = width * RATIO;
/* Handle if it's a "work" thing */
  if (!strcmp (o->name, "work")) {
    thisweek->work += (m2 - m1);
    d->work += (m2 - m1);
      sprintf (d->work_descr, "%s", o->comment);
/*
    if (strlen (d->work_descr) == 0)
      sprintf (d->work_descr, "%s", o->comment);
    else {
      strcat (d->work_descr, " / ");
      strcat (d->work_descr, o->comment);
    }
*/
    if (d->work_descr[strlen (d->work_descr) - 1] == '\n')
      d->work_descr[strlen (d->work_descr) - 1] = '\0';
    if ((same_day (o->start.d, d->date)) && (same_day (o->finish.d, d->date))) {
      fprintf (workfile, "%s %02d %s %02d:%02d - %02d:%02d %s",
	       d->date.dom, o->start.d.d, o->start.d.mon,
	       o->start.time.h, o->start.time.m, o->finish.time.h, o->finish.time.m, o->comment);
    }
    else if (same_day (o->start.d, d->date)) {
      fprintf (workfile, "%s %02d %s %02d:%02d - %02d %s %02d:%02d %s",
	       d->date.dom, o->start.d.d, o->start.d.mon,
	       o->start.time.h, o->start.time.m, o->finish.d.d, o->finish.d.mon, o->finish.time.h, o->finish.time.m, o->comment);
    }
    if (strlen (o->comment) == 0)
      fprintf (workfile, "\n");
  }
  /* Handle if it's a break */
  if (!strcmp(o->name, "break")) {
    thisweek->work -= (m2 - m1);
    d->work -= (m2 - m1);
  }
  if (width > 0)
    fprintf (html, "<img border=1 src=\"gif/%s.gif\" alt=\"%s\" width=\"%d\" height=\"3\">", o->name, o->comment, width);
  if ((o->next != NULL) && (same_day (o->next->start.d, d->date))) {
    //printf("\no->next=%s\n", o->next->name);
    if (!strcmp (o->name, o->next->name)) {
      // We'll add any gaps it needs...
      m1 = m2;
      if (same_day (o->next->start.d, d->date))
	m2 = o->next->start.time.h * 60 + o->next->start.time.m;
      else
	m2 = 24 * 60;
      width = m2 - m1;
      width = width * RATIO;
      if (width > 0)
	fprintf (html, "<img border=0 src=\"gif/trans.gif\" alt=\"\" width=\"%d\" height=\"3\">", width);
    }
    else
      fprintf (html, "<BR>\n");
  }
  else
    fprintf (html, "<BR>\n");
}

void show_sleep (struct sleep_struct *s, struct diary_struct *d)
{
  int width, m1, m2;

  strcat (summary, "");

// TRY
  if (same_day (s->try.d, d->date))
    m1 = s->try.time.h * 60 + s->try.time.m;
  else
    m1 = 0;
  if (same_day (s->sleep.d, d->date))
    m2 = s->sleep.time.h * 60 + s->sleep.time.m;
  else
    m2 = 24 * 60;
  width = m2 - m1;
  width = width * RATIO;

  if ((s->prev == NULL) && (m1 > 0)) {
    int w = m1 * RATIO;
    fprintf (html,
	     "<img border=0 src=\"gif/wake.gif\" alt=\"wake til %02d:%02d\" width=\"%d\" height=\"15\">",
	     s->try.time.h, s->try.time.m, w);
  }
  if ((width > 0)
      && ((same_day (s->try.d, d->date))
	  || (same_day (s->sleep.d, d->date)))) {
    if (same_day (s->try.d, d->date)) {
      sprintf (tmp_summ, "Try: %02d:%02d", s->try.time.h, s->try.time.m);
      strcat (summary, tmp_summ);
    }
    fprintf (html,
	     "<img border=1 src=\"gif/try.gif\" alt=\"try %02d:%02d to %02d:%02d\" width=\"%d\" height=\"15\">",
	     s->try.time.h, s->try.time.m, m2 / 60, m2 % 60, width);
  }
// SLEEP
  if (same_day (s->sleep.d, d->date)) {
    m1 = s->sleep.time.h * 60;
    m1 += s->sleep.time.m;
  }
  else
    m1 = 0;
  if (same_day (s->wake.d, d->date))
    m2 = s->wake.time.h * 60 + s->wake.time.m;
  else
    m2 = 24 * 60;
  width = m2 - m1;
  width = width * RATIO;
  if ((width > 0)
      && ((same_day (s->sleep.d, d->date))
	  || (same_day (s->wake.d, d->date)))) {
    if (same_day (s->sleep.d, d->date)) {
      sprintf (tmp_summ, " Sleep: %02d:%02d", s->sleep.time.h, s->sleep.time.m);
      sleep_today += (m2 - m1);
      strcat (summary, tmp_summ);
    }
    else {
      sleep_today += (m2 - m1);	// This should happen anyway!
    }
    fprintf (html,
	     "<img border=1 src=\"gif/sleep.gif\" alt=\"asleep %02d:%02d to %02d:%02d\" width=\"%d\" height=\"15\">",
	     s->sleep.time.h, s->sleep.time.m, m2 / 60, m2 % 60, width);
  }
// WAKE
  if (same_day (s->wake.d, d->date))
    m1 = s->wake.time.h * 60 + s->wake.time.m;
  else
    m1 = 0;
  if ((s->next != NULL) && (same_day (s->next->try.d, d->date)))
    m2 = s->next->try.time.h * 60 + s->next->try.time.m;
  else
    m2 = 24 * 60;
  width = m2 - m1;
  width = width * RATIO;
  if (width > 0)
    if (s->next != NULL) {
      if ((same_day (s->wake.d, d->date))
	  || (same_day (s->next->try.d, d->date))) {
	sprintf (tmp_summ, " Wake: %02d:%02d", s->wake.time.h, s->wake.time.m);
	strcat (summary, tmp_summ);
	fprintf (html,
		 "<img border=0 src=\"gif/wake.gif\" alt=\"awake %02d:%02d to %02d:%02d\" width=\"%d\" height=\"15\">",
		 s->wake.time.h, s->wake.time.m, m2 / 60, m2 % 60, width);
      }
    }
    else if (same_day (s->wake.d, d->date)) {
      sprintf (tmp_summ, " Wake: %02d:%02d", s->wake.time.h, s->wake.time.m);
      strcat (summary, tmp_summ);
      fprintf (html,
	       "<img border=0 src=\"gif/wake.gif\" alt=\"awake %02d:%02d to %02d:%02d\" width=\"%d\" height=\"15\">",
	       s->wake.time.h, s->wake.time.m, m2 / 60, m2 % 60, width);
    }
}

void get_bgcolour (char *colour, struct diary_struct *d)
{
  int weekend;

  if ((!strcmp (d->date.dom, "Sat")) || (!strcmp (d->date.dom, "Sun"))) {
    weekend = 1;
    sprintf (colour, " bgcolor=\"silver\"");
  }
  else {
    weekend = 0;
    sprintf (colour, " bgcolor=\"white\"");
  }

  if (strcmp (d->state, "normal")) {
    if (!strcmp (d->state, "training\n")) {
      sprintf (colour, " bgcolor=\"#dfada5\"");
    }
    if (!strcmp (d->state, "onsite\n")) {
      sprintf (colour, " bgcolor=\"#8cada5\"");
    }
    if (!strcmp (d->state, "holiday\n")) {
      if (weekend == 0)
	sprintf (colour, " bgcolor=\"#ffffaa\"");
      if (weekend == 1)
	sprintf (colour, " bgcolor=\"#ddddaa\"");
    }
  }

}

char *get_filename (struct week_struct *w)
{
  char *f = (char *) malloc (255);

  if (w == NULL)
    return " ";

  sprintf (f, "html/sleep-diary-wc-%s.%d.%s.html", w->first_day->date.dom, w->first_day->date.d, w->first_day->date.mon);

  if (short_output == 1)
    sprintf (f, "html/sleep-diary-summary.html");
  return f;
}

char *get_url (struct week_struct *w)
{
  char *f = (char *) malloc (255);

  if (w == NULL)
    return " ";

  sprintf (f, "sleep-diary-wc-%s.%d.%s.html", w->first_day->date.dom, w->first_day->date.d, w->first_day->date.mon);

  if (short_output == 1)
    sprintf (f, "sleep-diary-summary.html");
  return f;
}


void show_day (struct diary_struct *d, struct week_struct *w)
{
  struct other_struct *o;
  struct sleep_struct *s;
  char colour[100];
  char overall[100];
  char qos[100];
  int i = 0;

  sprintf (summary, "");

  get_bgcolour (colour, d);

  fprintf (html, "<tr valign=\"top\"%s>", colour);
  if ((short_output == 1) && (!strcmp (d->date.dom, "Mon"))) {
    fprintf (html,
	     "<th><a href=\"sleep-diary-wc-%s.%d.%s.html\">%s %d %s</a></th>\n",
	     d->date.dom, d->date.d, d->date.mon, d->date.dom, d->date.d, d->date.mon);
  }
  else
    fprintf (html, "<th>%s %d %s</th>\n", d->date.dom, d->date.d, d->date.mon);
  if (timebar == 1)
    fprintf (html, "    <td nowrap valign=\"top\" background=\"gif/timebar.gif\">");
  else {
    fprintf (html, "    <td nowrap valign=\"top\">");
    //fprintf (html, "<img src=\"gif/24.gif\" width=485 height=3><BR>");
  }
  if (work + details > 0) {
    o = d->other;
    while (o != NULL) {
      //printf ("%02d %s : %s %02d:%02d - %02d:%02d\n", d->date.d, d->date.mon, o->name,
      //o->start.time.h, o->start.time.m, o->finish.time.h, o->finish.time.h); 
      if (!strcmp (o->name, "work")) {
	if (work == 1) {
	  show_other (o, d);
	}
      }
      else {
	show_other (o, d);
      }
      o = o->next;
    }
/*
	FIXME
    if (d->work > 0) {
      fprintf (workfile, "%s %02d %s: 09:00 - %02d:%02d (%2dh %02dm) - %s\n", d->date.dom, d->date.d, d->date.mon, (d->work / 60) + 9, d->work % 60, d->work / 60, d->work % 60, d->work_descr);
    }
*/
  }
  s = d->sleep;
  while (s != NULL) {
    show_sleep (s, d);
    s = s->next;
  }

  show_sunrise (d);

  sprintf (overall, "%d", d->overall);
  sprintf (qos, "%d", d->qos);
  fprintf (html,
	   "<BR>\n<table width=\"100%\"><tr><td align=left><font size=-1>%s</font></td><td align=right><font size=-1>Total Sleep: %02d:%02d</font></td></tr></table>",
	   summary, sleep_today / 60, sleep_today % 60);
  thisweek->sleep+=sleep_today;
  fprintf (html, "\n    </td>\n");
  //fprintf (html, "<td>%d %d<img border=1 src=\"gif/mood/%s.gif\" width=\"28\" height=\"32\">", d->overall, d->qos, overall);
  fprintf (html, "<td><img border=1 src=\"gif/mood/%s.gif\" width=\"28\" height=\"32\">", overall);
  fprintf (html, "<img border=1 src=\"gif/qos/%s.gif\" width=\"28\" height=\"32\"><BR>%d,%d; %s</td>", qos, d->overall, d->qos,
	   d->feel);
  if ((short_output == 1) && (comments == 1)) {
    if (strlen (d->comments) > short_length)
      d->comments[short_length] = '\0';
  }
  if (comments == 1)
    fprintf (html, "<td>%s    </td>\n", d->comments);
  fprintf (html, "</tr>\n");
}

void show_weeks (struct week_struct *tw)
{
  struct week_struct *w = first_week;
  char this_month[25];
  sprintf (this_month, "%s", w->first_day->date.mon);

  fprintf (html, "\n<HR noshade>\n");
  fprintf (html, "\n<table border=1>\n");
  fprintf (html, "<tr><th>%s</th>", w->first_day->date.mon);
  while (w != NULL) {
    if (strcmp (this_month, w->first_day->date.mon))
      fprintf (html, "</tr>\n<tr><th>%s</th>", w->first_day->date.mon);
    sprintf (this_month, "%s", w->first_day->date.mon);
    fprintf (html, "\n<td>");
/*
    if (tw!=w) fprintf(html, "<a href=\"sleep-diary-wc-%s.%d.%s.html\">", w->first_day->date.dom, w->first_day->date.d, w->first_day->date.mon);
    else fprintf(html, "<B>");
    fprintf(html, "%s %d %s", w->first_day->date.dom, w->first_day->date.d, w->first_day->date.mon);
    if (tw!=w) fprintf(html, "</a>");
    else fprintf(html, "</B>");
*/
    fprintf (html, "<a href=\"sleep-diary-wc-%s.%d.%s.html\">",
	     w->first_day->date.dom, w->first_day->date.d, w->first_day->date.mon);
    fprintf (html, "%s %d %s", w->first_day->date.dom, w->first_day->date.d, w->first_day->date.mon);
    fprintf (html, "</a>");
    fprintf (html, "</td>");
    if (w->next == NULL)
      fprintf (html, " </tr> ");
    w = w->next;
  }
  fprintf (html, "\n</table>\n<HR noshade>\n");

}

void show_year_comments (FILE * html)
{
  FILE *infile;
  char s[255];

  infile = fopen ("year.html", "r");
  if (infile) {
    while (fgets (s, 255, infile)) {
      fprintf (html, "%s", s);
    }
    fclose (infile);
  }
}


void show_results (int idx)
{
  struct diary_struct *d = first_entry;
  char *filename;
  int w;
  FILE *key;
  char s[255];
  int i = 0;

  statsfile = fopen ("html/stats.html", "w");
  fprintf (statsfile, "<table border=1>\n");

  w = (24 * 60 * RATIO) + 10;

  thisweek = first_week;
  filename = get_filename (thisweek);
  html = fopen (filename, "w");
  if (short_output == 0)
    sprintf (s, "Sleep Diary (%s %d %s)", thisweek->first_day->date.dom,
	     thisweek->first_day->date.d, thisweek->first_day->date.mon);
  else
    sprintf (s, "Sleep Diary");
  fprintf (html, "<html>\n<head><title>%s</title></head>\n<body bgcolor=\"white\">\n", s);
  //show_weeks(thisweek);
  fprintf (html, "<center><a href=\"../\">Home</a> | <a href=\".\">Index and Legend</a>");
  if (short_output == 0) {
    fprintf (html, " | <a href=\"sleep-diary-summary.html\">Summary</a></center>\n");
    fprintf (html, "\n<a href=\"%s\">Next week</a>\n", get_url (thisweek->next));
  }
  else
    fprintf (html, "</center>\n");
  fprintf (html, "<table border=1>\n");
  fprintf (html,
	   "<th>Date</th><th width=%d>Sleep<BR><img width=%d src=\"gif/timeline.gif\"</th><th>Wake / QoS / Overall</th>", w, w);
  if (comments == 1)
    fprintf (html, "<th>Comments</th>\n");

  while (d != NULL) {
/*printf("date: %s %d %s\n", d->date.dom, d->date.d, d->date.mon);
printf("  %d %d %s %s\n", d->qos, d->overall, d->feel, d->comments);*/
    sleep_today = 0;
    if ((thisweek->next != NULL) && (d == thisweek->next->first_day)
	&& (short_output == 0)) {
      fprintf (html, "</table>\n");
      sort_stats (i);
      show_stats (i, thisweek);
      if (thisweek->work > 0) {
	fprintf (html, "<i>Total Work this week (target 37h:30m): %02dh:%02dm</i>\n", thisweek->work / 60, thisweek->work % 60);
	fprintf (html, "<BR><i>Total Sleep this week (8*7 = 56h 00m): %02dh:%02dm (avg %02dh:%02dm)</i>\n", thisweek->sleep / 60, thisweek->sleep % 60, (thisweek->sleep/7)/60, (thisweek->sleep/7)%60);
	fprintf (workfile, "Total Work this week : %02dh %02dm\n\n", thisweek->work / 60, thisweek->work % 60);
      }
      fprintf (html, "<center><a href=\"../\">Home</a> | <a href=\".\">Index and Legend</a>");
      if (short_output == 0)
	fprintf (html, " | <a href=\"sleep-diary-summary.html\">Summary</a>");
      fprintf (html, "</center>\n");
      //show_weeks(thisweek);
      fprintf (html, "</body>\n</html>\n");
      thisweek->next->prev = thisweek;
      thisweek = thisweek->next;
      fclose (html);
      filename = get_filename (thisweek);
      html = fopen (filename, "w");
      sprintf (s, "Sleep Diary (%s %d %s)",
	       thisweek->first_day->date.dom, thisweek->first_day->date.d, thisweek->first_day->date.mon);
      fprintf (html, "<html>\n<head><title>%s</title></head>\n<body bgcolor=\"white\">\n", s);
      //show_weeks(thisweek);
      fprintf (html, "<center><a href=\"../\">Home</a> | <a href=\".\">Index and Legend</a>");
      if (short_output == 0)
	fprintf (html, " | <a href=\"sleep-diary-summary.html\">Summary</a>");
      fprintf (html, "</center>\n");
      if (short_output == 0) {
	if (thisweek->prev)
	  fprintf (html, "\n<a href=\"%s\">Previous week </a>\n", get_url (thisweek->prev));
	if (thisweek->next)
	  fprintf (html, "\n<a href=\"%s\">Next week</a>\n", get_url (thisweek->next));
      }
      fprintf (html, "<table border=1>\n");
      fprintf (html,
	       "<th>Date</th><th width=%d>Sleep<BR><img width=%d src=\"gif/timeline.gif\"</th><th>Wake / QoS / Overall</th><th>Comments</th>\n",
	       w, w);
    }
    show_day (d, thisweek);
    i++;
    d = d->next;
  }
  fprintf (html, "</table>\n");
  if (short_output == 0) {
    sort_stats (i);
    show_stats (i, thisweek);
    fprintf (html, "<i>Total Work this week (target 37.5h): %02dh:%02dm</i>\n", thisweek->work / 60, thisweek->work % 60);
	fprintf (html, "<BR><i>Total Sleep this week (8*7 = 56h 00m): %02dh:%02dm (avg %02dh:%02dm)</i>\n", thisweek->sleep / 60, thisweek->sleep % 60, (thisweek->sleep/7)/60, (thisweek->sleep/7)%60);
    fprintf (workfile, "Total Work this week : %02dh %02dm\n", thisweek->work / 60, thisweek->work % 60);
    fprintf (html, "<center><a href=\"../\">Home</a> | <a href=\".\">Index and Legend</a>");
    fprintf (html, "</center>\n");
    fprintf (html, "</body>\n</html>\n");

    fclose (html);
  }

  html = fopen ("html/index.html", "w");
  fprintf (html, "<html>\n<head><title>Sleep Diary (Index)</title></head>\n<body bgcolor=\"white\">\n");
  fprintf (html, "<h1>Steve Parker's Sleep Diary - Index</h1>\n");
  fprintf (html, "<h3>Generated by <a href=\"http://steve-parker.org/sleepdiary/\">sleepdiary</a> (c) Steve Parker 2001</h3>\n");
  show_weeks (thisweek);
  show_year_comments (html);
  fprintf (html, "<HR noshade>\n");
  fprintf (html, "<a href=\"stats.html\">Stats for the year so far</a>\n");
  fprintf (html, "<HR noshade>\n");
  fprintf (html,
	   "Click here to see an all-time<a href=\"./sleep-diary-summary.html\">Summary view</a> - less detail, but the pattern is more visible.\n");
  sort_stats (idx);
  show_stats (idx, thisweek);
  key = fopen ("key.html", "r");
  while (fgets (s, 255, key))
    fprintf (html, "%s", s);
  fclose (key);
  fprintf (html, "</body>\n</html>\n");

  fclose (html);
  fclose (statsfile);

}

struct diary_struct *read_diary (char *s, char *t, char *u)
{
  int n;
  struct week_struct *w;
  struct diary_struct *d = (struct diary_struct *) malloc (sizeof (struct diary_struct));
  if (d==NULL) printf("Failed to malloc\n");

  if (sscanf (s, "%s %d %s %d %d %n", d->date.dom, &d->date.d, d->date.mon, &d->overall, &d->qos, &n) < 5) {
    printf ("Error parsing diary line: %s", s);
    return NULL;
  }
  if (n != strlen (s))
    sprintf (d->state, "%s", s + n);
  else
    sprintf (d->state, "normal");
  //printf("d->state=%s", d->state);
  d->feel = (char *) malloc (strlen (t) + 1);
  sprintf (d->feel, "%s", t);
  d->comments = (char *) malloc (strlen (u) + 1);
  sprintf (d->comments, "%s", u);
  d->other = NULL;
  d->next = NULL;
  if ((first_week == NULL) || (!strcmp (d->date.dom, "Mon"))) {
    if (first_week == NULL) {
      thisweek = (struct week_struct *) malloc (sizeof (struct week_struct));
      first_week = thisweek;
    }
    else {
      w = (struct week_struct *) malloc (sizeof (struct week_struct));
      thisweek->next = w;
      thisweek = w;
    }
    thisweek->first_day = d;
    thisweek->sleep=0;
  }

  return d;
}

int timediff (struct tod_struct from, struct tod_struct till, char *x)
{
  // return time in minutes between the two events...
  int res = 0;
  int h, m;
  int tth;
  struct diary_struct *d = first_entry;

  tth = till.time.h;

  h = till.time.h - from.time.h;
  m = till.time.m - from.time.m;

/*
  if (h<0) h+=24;
  if (m<0) h+=1;
*/
  if (!same_day (from.d, till.d)) {
    while (!same_day (d->date, from.d))
      d = d->next;
    while (!same_day (d->date, till.d)) {
      h += 24;
      d = d->next;
    }
  }

  res = (h * 60) + m;

/*
  printf("%-10s: %02d/%s @ %02d:%02d - %02d/%s @ %02d:%02d = %02d:%02d (%d)\n", x, from.d.d, from.d.mon, from.time.h, from.time.m, 
	till.d.d, till.d.mon, till.time.h, till.time.m, res/60, res%60, res);
*/

  return res;

}

struct sleep_struct *read_sleep (char *s, int idx, struct sleep_struct *prev)
{
  struct sleep_struct *sl = (struct sleep_struct *) malloc (sizeof (struct sleep_struct));
  char msg[255];
  int i;

  sprintf (msg, "foo");

  i = sscanf (s, "%d %s %d %d %d %s %d %d %d %s %d %d %s", &sl->try.d.d,
	      sl->try.d.mon, &sl->try.time.h, &sl->try.time.m,
	      &sl->sleep.d.d, sl->sleep.d.mon, &sl->sleep.time.h,
	      &sl->sleep.time.m, &sl->wake.d.d, sl->wake.d.mon, &sl->wake.time.h, &sl->wake.time.m, msg);


  /* Sometimes I've missed a few nights' entries... ignore anything I've been told to ignore */
  if (i > 12) {
    if (idx > 0) {
      if (!strcmp (msg, "missed_some")) {
	// printf ("Ignoring %d %s - claims %d hours awake\n", sl->try.d.d, sl->try.d.mon, (timediff (prev->wake, sl->try, "try") / 60));
	stats[ACT_WAKE][idx-1] = 0;
	stats[ACT_TRY][idx] = 0;
	stats[ACT_SLEEP][idx] = 0;
	return sl;
      }
    }
  }

  if (idx > 0)
    stats[ACT_WAKE][idx - 1] = timediff (prev->wake, sl->try, "wake");
  else
    stats[ACT_WAKE][idx] = (1440 - timediff (sl->try, sl->wake, "try"));
  stats[ACT_TRY][idx] = timediff (sl->try, sl->sleep, "try");
  stats[ACT_SLEEP][idx] = timediff (sl->sleep, sl->wake, "sleep");

  return sl;
}

struct other_struct *read_other (char *s)
{
  int res;
  char altstr[255];
  struct other_struct *o;

  o = (struct other_struct *) malloc (sizeof (struct other_struct));
  o->next = NULL;
  if (o == NULL) {
    printf ("Error in malloc()\n");
  }
  o->name = (char *) malloc (100);
  if (o->name == NULL)
    printf ("Error in malloc name\n");
  o->comment = (char *) malloc (200);
  if (o->comment == NULL)
    printf ("Error in malloc comment\n");

  res = sscanf (s, "%s %d %s %d %d %d %s %d %d %200c", o->name,
		&o->start.d.d, o->start.d.mon, &o->start.time.h,
		&o->start.time.m, &o->finish.d.d, o->finish.d.mon, &o->finish.time.h, &o->finish.time.m, o->comment);

  if ((res == 9) && (strcmp (o->name, "work"))) {
    sprintf (altstr, " %s: %02d:%02d to %02d:%02d", o->name,
	     o->start.time.h, o->start.time.m, o->finish.time.h, o->finish.time.m);
    o->comment = (char *) strdup (altstr);
  }
  else if (strcmp (o->name, "work")) {
    sprintf (altstr, "%s: %s", o->name, o->comment);
    sprintf (o->comment, "%s", altstr);
    // Now strip the trailing \n to keep the HTML tidy...
    o->comment[strlen (o->comment) - 1] = '\0';
    //strcat(o->comment, "</i>");
  }

  return o;
}


void position_sleep (struct sleep_struct *sl)
{
  struct diary_struct *d;

  d = first_entry;
  while (d != NULL) {
    /* Note: Yes, we can put a sleep entry into multiple diary entries. 
       That way, each diary entry can just ignore entries which don't refer to it. */

    if ((same_day (d->date, sl->try.d))
	|| (same_day (d->date, sl->sleep.d))
	|| (same_day (d->date, sl->wake.d))) {
      if (d->sleep == NULL)
	d->sleep = sl;
      else {
	struct sleep_struct *sl2 = d->sleep;
	while (sl2->next != NULL)
	  sl2 = sl2->next;
	sl2->next = sl;
	sl->prev = sl2;
      }
    }
    d = d->next;
  }
}

void position_other (struct other_struct *o)
{
  struct diary_struct *d;

  d = first_entry;

  while (d != NULL) {
    /* Note: Yes, we can put an entry into multiple diary entries. 
       That way, each diary entry can just ignore entries which don't refer to it. */

    if ((same_day (d->date, o->start.d)) || (same_day (d->date, o->finish.d))) {
      if (d->other == NULL) {
	d->other = o;
	//  printf("%d %s -> other = %s\n", d->date.d, d->date.mon, o->name);
      }
      else {
	struct other_struct *o2 = d->other;
	while (o2->next != NULL) {
	  //printf("%s %2d %s @ %02d:%02d\n", o2->name, o2->start.d.d, o2->start.d.mon, o2->start.time.h, o2->start.time.m);
	  o2 = o2->next;
	}
	if (o2 != o) {
	  // o2 could be o because of the way we're adding these regardless of the start date.
	  //printf ("%d %s : %s (%d) %02d:%02d -> next = %s (%d) %02d:%02d - %d\n", d->date.d, d->date.mon,
	  //o2->name, o2->start.d.d, o2->start.time.h, o2->start.time.m, o->name,
	  //o->start.d.d, o->start.time.h, o->start.time.m, o->finish.d.d);
	  o2->next = o;
	  o->prev = o2;
	}
      }
    }
    d = d->next;
  }
}

void get_options (int argc, char *argv[])
{
  int c;
  extern char *optarg;

  short_output = 0;
  details = 0;
  work = 0;
  short_length = 500;
  comments = 1;
  alcohol = 1;
  timebar = 0;

  while ((c = getopt (argc, argv, "cdsawvl:?")) != -1) {
    switch (c) {
    case 'c':
      comments = 0;
      break;
    case 'w':
      work = 1;
      break;
    case 'd':
      details = 1;
      break;
    case 'a':
      alcohol = 0;
      break;
    case 'l':
      short_length = atoi (optarg);
      break;
    case 's':
      short_output = 1;
      break;
    case 'v':
      timebar = 1;
      break;
    case '?':
      printf ("\nUsage: sleep [ -s [ -d [ -l length ] ] -c ]\n\n");
      printf ("      Where -s selects the summary output\n");
      printf ("            -d adds details (plus work, drink, etc)\n");
      printf ("            -c *disables* comments\n");
      printf ("	      -l specifies the length of comments (in characters) to add\n");
      printf ("            -a doesn't show alcohol\n");
      exit (1);
      break;
    }
  }
  if (short_output == 0) {
    details = 1;
    comments = 1;
  }
  if (details == 1) {
    work = 1;
    //anything else specific = 1
  }
}


int main (int argc, char *argv[])
{
  FILE *sleep_dat;
  FILE *other_dat;
  FILE *diary_dat;
  char s[1024];
  char t[1024];
  char u[1024];
  struct diary_struct *d = NULL;
  struct diary_struct *prev_d = NULL;
  struct sleep_struct *sl = NULL;
  struct other_struct *o = NULL;
  int idx = 0;

  workfile = fopen ("work.txt", "w");

  get_options (argc, argv);

  diary_dat = fopen ("dat/diary.dat", "r");
  while (fgets (s, 1024, diary_dat)) {
    if (s[0] != '#') {
      fgets (t, 1024, diary_dat);
      fgets (u, 1024, diary_dat);
      d = read_diary (s, t, u);
      if (first_entry == NULL) {
	first_entry = d;
      }
      d->prev = prev_d;
      if (prev_d != NULL)
	prev_d->next = d;
      d->next = NULL;
      prev_d = d;
    }
  }
  fclose (diary_dat);
  sleep_dat = fopen ("dat/sleep.dat", "r");
  while (fgets (s, 1024, sleep_dat)) {
    sl = read_sleep (s, idx, sl);
    position_sleep (sl);
    idx++;
  }
  fclose (sleep_dat);

  other_dat = fopen ("dat/other.dat", "r");
  while (fgets (s, 1024, other_dat)) {
    o = read_other (s);
    position_other (o);
  }
  fclose (other_dat);


  other_dat = fopen ("dat/sunrise.dat", "r");
  while (fgets (s, 1024, other_dat)) {
    o = read_other (s);
    position_other (o);
  }
  fclose (other_dat);


  show_results (idx);

  fclose (workfile);
  return 0;
}
